/* ACR2000 Kernel Driver
 * Copyright (C) 1998 George Lebl
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 * This driver was done using the excellent Linux Device Drivers book
 * written by Alessandro Rubini as a guide.
 *
 * This driver implements a "protocol" you can use to talk to the
 * ACR2000 card for controlling steppers and servos, right now it's
 * using a device with major number of 127 (you can change that and
 * the addresses below). It should also work with acr8000 I'd guess.
 *
 * This could all really be done in user-space I guess, but this seemed
 * like the more reliable way.
 *
 * v0.2 - made variables for configuration with insmod
 *        export no symbols
 *        took out debugging info
 *
 * v0.1 - Initial Release
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>

#define ACR2000_DEFAULT_ADDR		0x300
#define ACR2000_DEFUALT_MAJOR_DEV_NUM	127

#define ACR2000_RANGE			0x004
#define ACR2000_STATUS_BYTE_OFFSET	0x002


int io = ACR2000_DEFAULT_ADDR;
int major_num = 127;

#ifndef NULL
#define NULL ((void *)0)
#endif

static int
acr2000_detect(unsigned int port, unsigned int range)
{
	int err;

	if((err=check_region(port,range)) < 0)
		return err;

	request_region(port,range,"acr2000");

	return 0;
}

static void
acr2000_release(unsigned int port, unsigned int range)
{
	release_region(port,range);
}

static int
acr2000_read(struct inode *inode, struct file *file, char *s, int size)
{
	unsigned in;
	unsigned long flags;
	int i;
	char *p;

	if(size<=0)
		return 0;


	p = kmalloc(size,GFP_KERNEL);

	for(i=0;i<size;i++) {
		
		save_flags(flags);
		cli();

		in = inb_p(io + ACR2000_STATUS_BYTE_OFFSET);

		if(!(in & 0x20)) {
			restore_flags(flags);
			break;
		}
		in = inb_p(io);
		restore_flags(flags);
		p[i] = in;
	}
	if(i>0)
		memcpy_tofs(s,p,i);
	kfree(p);
	return i;
}

static int
acr2000_write(struct inode *inode, struct file *file, const char *s, int size)
{
	unsigned in;
	unsigned long flags;
	char *p;
	int i;

	if(size<=0)
		return 0;

	p = kmalloc(size,GFP_KERNEL);
	memcpy_fromfs(p,s,size);

	for(i=0;i<size;i++) {

		save_flags(flags);
		cli();

		in = inb_p(io + ACR2000_STATUS_BYTE_OFFSET);

		if(!(in & 0x80)) {
			restore_flags(flags);
			break;
		}
		outb_p(p[i],io+1);

		restore_flags(flags);

	}
	kfree(p);
	return i;
}

struct file_operations acr2000_fops = {
	NULL, /*lseek*/
	acr2000_read,
	acr2000_write,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
	
};


static int
acr2000_register_dev(void)
{
	int err;

	if((err=register_chrdev(major_num,"acr2000",&acr2000_fops)) < 0)
		return err;
	return 0;
}

static void
acr2000_unregister_dev(void)
{
	unregister_chrdev(major_num,"acr2000");
}

int
init_module(void)
{
	int err;

	register_symtab(NULL);

	err = acr2000_detect(io,ACR2000_RANGE);
	if(err) goto return_error;

	err = acr2000_register_dev();
	if(err) goto unreg_port;

	return 0;

	unreg_port: acr2000_release(io,ACR2000_RANGE);
	return_error: return err;
}

void
cleanup_module(void)
{

	acr2000_unregister_dev();
	acr2000_release(io,ACR2000_RANGE);

}
